#include "entity_health.hpp"
#include "unit.hpp"
#include "utils/godot_macros.hpp"

void EntityHealth::_bind_methods() {
#define CLASSNAME EntityHealth
    GDFUNCTION_ARGS(damage, "amount");
    GDFUNCTION_ARGS(damaged_by, "amount", "source");
    GDFUNCTION_ARGS(healed_by, "amount", "source");
    GDPROPERTY(injury_max, gd::Variant::INT);
    GDPROPERTY(wounds_max, gd::Variant::INT);

    GDSIGNAL("damage",
            gd::PropertyInfo(gd::Variant::INT, "remaining"),
            gd::PropertyInfo(gd::Variant::INT, "change"),
            gd::PropertyInfo(gd::Variant::OBJECT, "source", gd::PROPERTY_HINT_NODE_TYPE, "Unit"));
    GDSIGNAL("heal",
            gd::PropertyInfo(gd::Variant::INT, "remaining"),
            gd::PropertyInfo(gd::Variant::INT, "change"),
            gd::PropertyInfo(gd::Variant::OBJECT, "source", gd::PROPERTY_HINT_NODE_TYPE, "Unit"));
    GDSIGNAL("health_changed",
            gd::PropertyInfo(gd::Variant::INT, "remaining"),
            gd::PropertyInfo(gd::Variant::INT, "change"));
    GDSIGNAL("death", gd::PropertyInfo(gd::Variant::OBJECT, "source", gd::PROPERTY_HINT_NODE_TYPE, "Unit"));
}

void EntityHealth::_enter_tree() {
    this->injury_current = this->injury_max;
    this->wounds_current = this->wounds_max;
    this->emit_signal("damage", this->injury_current, 0, nullptr);
}

void EntityHealth::damage(int amount) {
    this->damaged_by(amount, nullptr);
}

void EntityHealth::damaged_by(int amount, Unit *source) {
    amount = gd::Math::abs(amount);
    if(this->injury_current <= 0) return; // do not take damage when already dead
    this->injury_current -= amount;
    this->emit_signal("damage", this->injury_current, amount, source);
    this->emit_signal("health_changed", this->injury_current, -amount);
    if(this->injury_current <= 0) {
        this->emit_signal("death", source);
    }
}

void EntityHealth::healed_by(int amount, Unit *source) {
    amount = gd::Math::abs(amount);
    this->injury_current = gd::Math::min(this->injury_max, this->injury_current + amount);
    this->emit_signal("heal", this->injury_current, amount, source);
    this->emit_signal("health_changed", this->injury_current, amount);
}

bool EntityHealth::get_is_dead() const {
    return this->injury_current <= 0;
}

void EntityHealth::set_injury_max(int max_health) {
    this->injury_max = max_health;
}

int EntityHealth::get_injury_max() const {
    return this->injury_max;
}

int EntityHealth::get_injury_current() const {
    return this->injury_current;
}

void EntityHealth::set_wounds_max(int max_health) {
    this->wounds_max = max_health;
}

int EntityHealth::get_wounds_max() const {
    return this->wounds_max;
}

int EntityHealth::get_wounds_current() const {
    return this->wounds_current;
}
